"""Test cases for environment lock management."""

import shutil
import tempfile
import tomllib

import pytest

from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, cast

from pytest_subtests import SubTests

from venvstacks.stacks import (
    BuildEnvError,
    BuildEnvironment,
    EnvironmentLock,
    LayerBaseName,
    LayerEnvBase,
    LayerVariants,
    StackSpec,
    _hash_strings,
)


@pytest.fixture
def temp_dir_path() -> Generator[Path, None, None]:
    with tempfile.TemporaryDirectory() as dir_name:
        yield Path(dir_name)


##############################
# EnvironmentLock test cases
##############################


def test_default_state(temp_dir_path: Path) -> None:
    req_path = temp_dir_path / "requirements.txt"
    env_lock = EnvironmentLock(req_path, (), ())
    # Declared requirements file is only written when requested
    assert env_lock.declared_requirements == ()
    assert env_lock._lock_input_path == temp_dir_path / "requirements.in"
    assert not env_lock.locked_requirements_path.exists()
    no_dependencies_hash = env_lock._lock_input_hash
    assert no_dependencies_hash is not None
    env_lock.prepare_lock_inputs()
    assert env_lock._lock_input_path.read_text() != ""
    assert env_lock._lock_input_hash == no_dependencies_hash
    # Locked requirements file must be written externally
    assert env_lock.locked_requirements_path == req_path
    assert not env_lock.locked_requirements_path.exists()
    assert env_lock._requirements_hash is None
    # Metadata file is only written when requested
    assert env_lock._lock_metadata_path == temp_dir_path / "requirements.txt.json"
    assert not env_lock._lock_metadata_path.exists()
    assert env_lock.load_valid_metadata() is None


def test_load_with_consistent_file_hashes(temp_dir_path: Path) -> None:
    req_path = temp_dir_path / "requirements.txt"
    env_lock = EnvironmentLock(req_path, (), ())
    env_lock.prepare_lock_inputs()
    env_lock.locked_requirements_path.write_text("")
    env_lock.update_lock_metadata()
    assert env_lock._lock_input_hash is not None
    assert env_lock._requirements_hash is not None
    env_lock_metadata = env_lock.load_valid_metadata()
    assert env_lock_metadata is not None
    # Loading the lock without changes gives the same metadata
    loaded_lock = EnvironmentLock(req_path, (), ())
    assert loaded_lock._lock_input_hash == env_lock._lock_input_hash
    assert loaded_lock._requirements_hash == env_lock._requirements_hash
    assert loaded_lock.load_valid_metadata() == env_lock_metadata


def test_load_with_inconsistent_input_hash(temp_dir_path: Path) -> None:
    req_path = temp_dir_path / "requirements.txt"
    env_lock = EnvironmentLock(req_path, (), ())
    env_lock.prepare_lock_inputs()
    env_lock.locked_requirements_path.write_text("Hash fodder")
    env_lock.update_lock_metadata()
    assert env_lock._lock_input_hash is not None
    assert env_lock._requirements_hash is not None
    assert env_lock.load_valid_metadata() is not None
    # Loading the lock with different requirements invalidates the metadata
    loaded_lock = EnvironmentLock(req_path, ("some-requirement",), ())
    assert loaded_lock._lock_input_hash != env_lock._lock_input_hash
    assert loaded_lock._requirements_hash is None
    assert loaded_lock.load_valid_metadata() is None


def test_load_with_inconsistent_output_hash(temp_dir_path: Path) -> None:
    req_path = temp_dir_path / "requirements.txt"
    env_lock = EnvironmentLock(req_path, (), ())
    env_lock.prepare_lock_inputs()
    env_lock.locked_requirements_path.write_text("Hash fodder")
    env_lock.update_lock_metadata()
    assert env_lock._lock_input_hash is not None
    assert env_lock._requirements_hash is not None
    assert env_lock.load_valid_metadata() is not None
    # Loading the lock with a different lock file invalidates the metadata
    env_lock.locked_requirements_path.write_text("")
    loaded_lock = EnvironmentLock(req_path, (), ())
    assert loaded_lock._lock_input_hash == env_lock._lock_input_hash
    assert loaded_lock._requirements_hash != env_lock._requirements_hash
    assert loaded_lock.load_valid_metadata() is None


def test_requirements_file_hashing(temp_dir_path: Path) -> None:
    messy_requirements = [
        "# File header comment",
        "b==2.3.4  # Trailing comment",
        "    c==3.4.5  # Leading whitespace",
        "a==1.2.3  # Entry out of order",
        "",
        "# Preceding line intentionally blank",
        "d==4.5.6",
    ]
    clean_requirements = EnvironmentLock._clean_reqs(messy_requirements)
    expected_requirements = [
        "a==1.2.3",
        "b==2.3.4",
        "c==3.4.5",
        "d==4.5.6",
    ]
    assert clean_requirements == expected_requirements
    expected_hash = _hash_strings(expected_requirements)
    assert EnvironmentLock._hash_reqs(messy_requirements) == expected_hash
    req_input_path = temp_dir_path / "requirements.in"
    req_input_path.write_text("\n".join(messy_requirements))
    req_file_hash = EnvironmentLock._hash_req_file(req_input_path)
    assert req_file_hash == expected_hash


##################################
# Layer specification test cases
##################################

EMPTY_SCRIPT_PATH = Path(__file__).parent / "minimal_project/empty.py"
EXAMPLE_STACK_SPEC = """\
[[runtimes]]
name = "cpython-to-be-modified"
python_implementation = "cpython@3.11.11"
requirements = []

[[runtimes]]
name = "cpython-same-major-version-unaffected"
python_implementation = "cpython@3.11.9"
requirements = []

[[runtimes]]
name = "cpython-other-version-unaffected"
python_implementation = "cpython@3.12.9"
requirements = []

[[frameworks]]
name = "to-be-modified"
runtime = "cpython-to-be-modified"
requirements = []

[[frameworks]]
name = "other-app-dependency"
runtime = "cpython-to-be-modified"
requirements = []

[[frameworks]]
# Must be after "other-app-dependency" to allow the declared
# dependency to be switched without reordering the list
name = "dependent"
frameworks = ["to-be-modified"]
requirements = []

[[frameworks]]
name = "unaffected"
runtime = "cpython-other-version-unaffected"
requirements = []

[[applications]]
name = "to-be-modified"
launch_module = "launch.py"
# "dependent" must be first here to allow linearisation
# when that layer also depends on "other-app-dependency"
frameworks = ["dependent", "other-app-dependency"]
requirements = []

[[applications]]
name = "runtime-only"
runtime = "cpython-to-be-modified"
launch_module = "launch2.py"
requirements = []

[[applications]]
name = "unaffected"
launch_module = "launch2.py"
frameworks = ["unaffected"]
requirements = []
"""

EXPECTED_LAYER_NAMES = (
    "cpython-to-be-modified",
    "cpython-same-major-version-unaffected",
    "cpython-other-version-unaffected",
    "framework-to-be-modified",
    "framework-other-app-dependency",
    "framework-dependent",
    "framework-unaffected",
    "app-to-be-modified",
    "app-runtime-only",
    "app-unaffected",
)


def _define_lock_testing_env(
    spec_path: Path, spec_data: dict[str, Any]
) -> BuildEnvironment:
    stack_spec = StackSpec.from_dict(spec_path, spec_data)
    return stack_spec.define_build_environment()


def _partition_envs(build_env: BuildEnvironment) -> tuple[set[str], set[str]]:
    valid_locks: set[str] = set()
    invalid_locks: set[str] = set()
    for env in build_env.all_environments():
        set_to_update = invalid_locks if env.needs_lock() else valid_locks
        set_to_update.add(env.env_name)
    return valid_locks, invalid_locks


@contextmanager
def _modified_file(file_path: Path, contents: str) -> Generator[Any, None, None]:
    backup_path = file_path.rename(file_path.with_suffix(".bak"))
    if not contents.endswith("\n"):
        contents += "\n"
    try:
        file_path.write_text(contents)
        yield
    finally:
        file_path.unlink()
        backup_path.rename(file_path)


def test_build_env_layer_locks(temp_dir_path: Path, subtests: SubTests) -> None:
    # Built as a monolithic tests with subtests for performance reasons
    # (initial setup takes ~10 seconds, subsequent checks are fractions of a second)
    launch_module_path = temp_dir_path / "launch.py"
    shutil.copyfile(EMPTY_SCRIPT_PATH, launch_module_path)
    shutil.copyfile(EMPTY_SCRIPT_PATH, temp_dir_path / "launch2.py")
    spec_path = temp_dir_path / "venvstacks.toml"
    updated_spec_path = temp_dir_path / "venvstacks_updated.toml"
    spec_data = tomllib.loads(EXAMPLE_STACK_SPEC)
    build_env_to_lock = _define_lock_testing_env(spec_path, spec_data)
    # Check for divergence between stack spec and the expected results
    # This also keeps the test from trivially passing due to bugs in the iterators
    layer_names = tuple(env.env_name for env in build_env_to_lock.all_environments())
    assert layer_names == EXPECTED_LAYER_NAMES
    # Preliminary checks that locking the stack updates the state as expected
    assert build_env_to_lock._needs_lock()
    all_layer_names = {*EXPECTED_LAYER_NAMES}
    valid_locks, invalid_locks = _partition_envs(build_env_to_lock)
    assert valid_locks == set()
    assert invalid_locks == all_layer_names
    assert all(env.needs_lock() for env in build_env_to_lock.all_environments())

    # Check lock input file determination
    layers_to_lock_names: list[str] = []
    for runtime_env in build_env_to_lock.runtimes_to_lock():
        # Runtime environments are always ready to be locked
        assert runtime_env.kind == LayerVariants.RUNTIME
        layers_to_lock_names.append(runtime_env.env_name)
        runtime_lock_inputs = runtime_env.get_lock_inputs()
        expected_runtime_lock_inputs = (
            runtime_env.requirements_path,
            runtime_env.requirements_path.with_suffix(".in"),
            cast(list[Path], []),
        )
        assert runtime_lock_inputs == expected_runtime_lock_inputs
    for unlocked_env in build_env_to_lock.environments_to_lock():
        if unlocked_env.kind == LayerVariants.RUNTIME:
            continue
        # Layered environments can only be locked after the layers they depend on
        layers_to_lock_names.append(unlocked_env.env_name)
        with pytest.raises(BuildEnvError, match="unlocked dependencies"):
            unlocked_env.get_lock_inputs()
    # Ensure this check can't trivially pass due to bugs in the iterators
    assert tuple(layers_to_lock_names) == EXPECTED_LAYER_NAMES

    # Actually lock the environments
    build_env_to_lock.lock_environments()
    assert not build_env_to_lock._needs_lock()
    valid_locks, invalid_locks = _partition_envs(build_env_to_lock)
    assert valid_locks == all_layer_names
    assert invalid_locks == set()

    # Now check various modified stacks in the same folder as the locked stack
    # Ensure the expected layers are detected as no longer having valid locks
    # Note: modified stacks are never locked, so the nominal deps don't really matter
    unaffected_layer_names = {
        name for name in EXPECTED_LAYER_NAMES if name.endswith("-unaffected")
    }
    subtests_started = subtests_passed = 0  # Track subtest failures
    with subtests.test("Already locked stack with no changes"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == all_layer_names
        assert invalid_locks == set()
        assert not build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Enable implicit layer versioning at runtime layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["runtimes"][0]
        assert env_spec_to_modify["name"] == "cpython-to-be-modified"
        env_spec_to_modify["versioned"] = True
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Enable implicit layer versioning at framework layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["frameworks"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["versioned"] = True
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names | {
            "cpython-to-be-modified",
            "framework-other-app-dependency",
            "app-runtime-only",
        }
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Enable implicit layer versioning at application layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["applications"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["versioned"] = True
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_invalid_locks = {"app-to-be-modified"}
        expected_valid_locks = all_layer_names - expected_invalid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change declared requirements at runtime layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["runtimes"][0]
        assert env_spec_to_modify["name"] == "cpython-to-be-modified"
        env_spec_to_modify["requirements"] = ["pip==25.1"]
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change declared requirements at framework layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["frameworks"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["requirements"] = ["pip==25.1"]
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names | {
            "cpython-to-be-modified",
            "framework-other-app-dependency",
            "app-runtime-only",
        }
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change declared requirements at application layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["applications"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["requirements"] = ["pip==25.1"]
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_invalid_locks = {"app-to-be-modified"}
        expected_valid_locks = all_layer_names - expected_invalid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Switch Python maintenance release at runtime layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["runtimes"][0]
        assert env_spec_to_modify["name"] == "cpython-to-be-modified"
        assert env_spec_to_modify["python_implementation"].startswith("cpython@3.11.")
        updated_py_version = "cpython@3.11.5"
        assert env_spec_to_modify["python_implementation"] != updated_py_version
        env_spec_to_modify["python_implementation"] = updated_py_version
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == all_layer_names
        assert invalid_locks == set()
        assert not build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Update to new major Python version at runtime layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["runtimes"][0]
        assert env_spec_to_modify["name"] == "cpython-to-be-modified"
        assert env_spec_to_modify["python_implementation"].startswith("cpython@3.11.")
        env_spec_to_modify["python_implementation"] = "cpython@3.12.5"
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change declared runtime at framework layer"):
        # Even if the major Python version doesn't change, the layer needs updating
        # This is due to the relative path to the runtime saved in the deployed layer config
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        for env_spec_to_modify in spec_data_to_check["frameworks"]:
            # Must modify both framework layers, otherwise stack spec is rejected as inconsistent
            if env_spec_to_modify["name"] not in (
                "to-be-modified",
                "other-app-dependency",
            ):
                continue
            env_spec_to_modify["runtime"] = "cpython-same-major-version-unaffected"
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names | {
            "cpython-to-be-modified",
            "app-runtime-only",
        }
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change declared runtime at application layer"):
        # Even if the major Python version doesn't change, the layer needs updating
        # This is due to the relative base runtime path saved in the deployed layer config
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["applications"][1]
        assert env_spec_to_modify["name"] == "runtime-only"
        env_spec_to_modify["runtime"] = "cpython-same-major-version-unaffected"
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_invalid_locks = {"app-runtime-only"}
        expected_valid_locks = all_layer_names - expected_invalid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change framework dependencies at framework layer"):
        # Even if the locked requirements don't change, the layer needs updating
        # This is due to the relative framework paths saved in the deployed layer config
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["frameworks"][2]
        assert env_spec_to_modify["name"] == "dependent"
        assert env_spec_to_modify["frameworks"] == ["to-be-modified"]
        env_spec_to_modify["frameworks"] = ["other-app-dependency"]
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_valid_locks = unaffected_layer_names | {
            "cpython-to-be-modified",
            "framework-to-be-modified",
            "framework-other-app-dependency",
            "app-runtime-only",
        }
        expected_invalid_locks = all_layer_names - expected_valid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change framework dependencies at application layer"):
        # Even if the locked requirements don't change, the layer needs updating
        # This is due to the relative framework paths saved in the deployed layer config
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["applications"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["frameworks"].remove("dependent")
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_invalid_locks = {"app-to-be-modified"}
        expected_valid_locks = all_layer_names - expected_invalid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change launch module name at application layer"):
        # Even if the locked requirements don't change, the layer needs updating
        # This is due to the launch module needing to be invoked differently
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_spec_to_modify = spec_data_to_check["applications"][0]
        assert env_spec_to_modify["name"] == "to-be-modified"
        env_spec_to_modify["launch_module"] = "launch2.py"
        build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
        expected_invalid_locks = {"app-to-be-modified"}
        expected_valid_locks = all_layer_names - expected_invalid_locks
        valid_locks, invalid_locks = _partition_envs(build_env)
        assert valid_locks == expected_valid_locks
        assert invalid_locks == expected_invalid_locks
        assert build_env._needs_lock()
        subtests_passed += 1
    # Remaining subtests need to modify the actual envs, not just the layer specifications
    env_to_modify: LayerEnvBase
    with subtests.test("Change locked requirements at runtime layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_to_modify = build_env_to_lock.runtimes[
            LayerBaseName("cpython-to-be-modified")
        ]
        with _modified_file(
            env_to_modify.env_lock.locked_requirements_path, "pip==25.1"
        ):
            build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
            expected_valid_locks = unaffected_layer_names
            expected_invalid_locks = all_layer_names - expected_valid_locks
            valid_locks, invalid_locks = _partition_envs(build_env)
            assert valid_locks == expected_valid_locks
            assert invalid_locks == expected_invalid_locks
            assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change locked requirements at framework layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_to_modify = build_env_to_lock.frameworks[LayerBaseName("to-be-modified")]
        with _modified_file(
            env_to_modify.env_lock.locked_requirements_path, "pip==25.1"
        ):
            build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
            expected_valid_locks = unaffected_layer_names | {
                "cpython-to-be-modified",
                "framework-other-app-dependency",
                "app-runtime-only",
            }
            expected_invalid_locks = all_layer_names - expected_valid_locks
            valid_locks, invalid_locks = _partition_envs(build_env)
            assert valid_locks == expected_valid_locks
            assert invalid_locks == expected_invalid_locks
            assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change locked requirements at application layer"):
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        env_to_modify = build_env_to_lock.applications[LayerBaseName("to-be-modified")]
        with _modified_file(
            env_to_modify.env_lock.locked_requirements_path, "pip==25.1"
        ):
            build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
            expected_invalid_locks = {"app-to-be-modified"}
            expected_valid_locks = all_layer_names - expected_invalid_locks
            valid_locks, invalid_locks = _partition_envs(build_env)
            assert valid_locks == expected_valid_locks
            assert invalid_locks == expected_invalid_locks
            assert build_env._needs_lock()
        subtests_passed += 1
    with subtests.test("Change launch module content at application layer"):
        # Even if the locked requirements don't change, the layer needs updating
        # This is due to the launch module needing to be invoked differently
        subtests_started += 1
        spec_data_to_check = tomllib.loads(EXAMPLE_STACK_SPEC)
        with _modified_file(
            launch_module_path, "# Changed launch module contents"
        ):
            build_env = _define_lock_testing_env(updated_spec_path, spec_data_to_check)
            expected_invalid_locks = {"app-to-be-modified"}
            expected_valid_locks = all_layer_names - expected_invalid_locks
            valid_locks, invalid_locks = _partition_envs(build_env)
            assert valid_locks == expected_valid_locks
            assert invalid_locks == expected_invalid_locks
            assert build_env._needs_lock()
        subtests_passed += 1

    # Work around pytest-subtests not failing the test case when subtests fail
    # https://github.com/pytest-dev/pytest-subtests/issues/76
    assert subtests_passed == subtests_started, (
        f"Fail due to failed subtest(s) ({subtests_passed} < {subtests_started})"
    )



"""Test building the minimal project produces the expected results."""

import json
import shutil
import sys
import tempfile

from datetime import datetime, timezone
from pathlib import Path
from typing import Any, cast

# Use unittest for consistency with test_sample_project (which needs the better diff support)
import unittest
from unittest.mock import Mock

import pytest  # To mark slow test cases

from support import (
    ApplicationEnvSummary,
    DeploymentTestCase,
    EnvSummary,
    LayeredEnvSummary,
    ManifestData,
    SpecLoadingTestCase,
    make_mock_index_config,
    get_sys_path,
)

from venvstacks.stacks import (
    ArchiveBuildMetadata,
    ArchiveMetadata,
    StackPublishingRequest,
    BuildEnvironment,
    EnvNameBuild,
    EnvNameDeploy,
    StackSpec,
    PackageIndexConfig,
    PublishedArchivePaths,
    get_build_platform,
)
from venvstacks._util import get_env_python, capture_python_output, WINDOWS_BUILD

##################################
# Minimal project test helpers
##################################

_THIS_PATH = Path(__file__)
MINIMAL_PROJECT_PATH = _THIS_PATH.parent / "minimal_project"
MINIMAL_PROJECT_STACK_SPEC_PATH = MINIMAL_PROJECT_PATH / "venvstacks.toml"
MINIMAL_PROJECT_PATHS = (
    MINIMAL_PROJECT_STACK_SPEC_PATH,
    MINIMAL_PROJECT_PATH / "empty.py",
)


def _copy_project_input_files(working_path: Path) -> None:
    """Copy the input files for the project to the given temporary folder."""
    # To avoid side effects from lock file creation, copy project files to the working path
    for src_path in MINIMAL_PROJECT_PATHS:
        dest_path = working_path / src_path.name
        shutil.copyfile(src_path, dest_path)


def _define_build_env(working_path: Path) -> BuildEnvironment:
    """Define a build environment for the sample project in a temporary folder."""
    _copy_project_input_files(working_path)
    # Include "/../" in the spec path in order to test relative path resolution when
    # accessing the Python executables (that can be temperamental, especially on macOS).
    # The subdirectory won't be used for anything, so it being missing shouldn't matter.
    working_spec_path = working_path / "_unused_dir/../venvstacks.toml"
    stack_spec = StackSpec.load(working_spec_path)
    build_path = working_path / "_build🐸"
    return stack_spec.define_build_environment(build_path)


##################################
# Expected stack definitions
##################################

EXPECTED_RUNTIMES = [
    EnvSummary("cpython-3.11", ""),
]

EXPECTED_FRAMEWORKS = [
    LayeredEnvSummary("layerA", "framework-", "cpython-3.11", ()),
    LayeredEnvSummary("layerB", "framework-", "cpython-3.11", ("layerA",)),
    LayeredEnvSummary("layerC", "framework-", "cpython-3.11", ("layerA",)),
    LayeredEnvSummary(
        "layerD", "framework-", "cpython-3.11", ("layerB", "layerC", "layerA")
    ),
    LayeredEnvSummary("layerE", "framework-", "cpython-3.11", ("layerB", "layerA")),
    LayeredEnvSummary(
        "layerF", "framework-", "cpython-3.11", ("layerE", "layerB", "layerA")
    ),
]

EXPECTED_APPLICATIONS = [
    ApplicationEnvSummary(
        "empty",
        "app-",
        "cpython-3.11",
        ("layerD", "layerF", "layerE", "layerB", "layerC", "layerA"),
    ),
    ApplicationEnvSummary("no-framework", "app-", "cpython-3.11", ()),
]

EXPECTED_ENVIRONMENTS = EXPECTED_RUNTIMES.copy()
EXPECTED_ENVIRONMENTS.extend(EXPECTED_FRAMEWORKS)
EXPECTED_ENVIRONMENTS.extend(EXPECTED_APPLICATIONS)

# The expected manifest here omits all content dependent fields
# (those are checked when testing the full sample project)
ArchiveSummary = dict[str, Any]
ArchiveSummaries = dict[str, list[ArchiveSummary]]
BuildManifest = dict[str, ArchiveSummaries]
ARCHIVE_SUFFIX = ".zip" if WINDOWS_BUILD else ".tar.xz"
BUILD_PLATFORM = str(get_build_platform())
EXPECTED_MANIFEST: BuildManifest = {
    "layers": {
        "applications": [
            {
                "app_launch_module": "empty",
                "archive_build": 1,
                "install_target": "app-empty",
                "archive_name": f"app-empty{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerD",
                    "framework-layerF",
                    "framework-layerE",
                    "framework-layerB",
                    "framework-layerC",
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "app_launch_module": "empty",
                "archive_build": 1,
                "install_target": "app-no-framework",
                "archive_name": f"app-no-framework{ARCHIVE_SUFFIX}",
                "required_layers": [],
                "target_platform": BUILD_PLATFORM,
            },
        ],
        "frameworks": [
            {
                "archive_build": 1,
                "install_target": "framework-layerA",
                "archive_name": f"framework-layerA{ARCHIVE_SUFFIX}",
                "required_layers": [],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "archive_build": 1,
                "install_target": "framework-layerB",
                "archive_name": f"framework-layerB{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "archive_build": 1,
                "install_target": "framework-layerC",
                "archive_name": f"framework-layerC{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "archive_build": 1,
                "install_target": "framework-layerD",
                "archive_name": f"framework-layerD{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerB",
                    "framework-layerC",
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "archive_build": 1,
                "install_target": "framework-layerE",
                "archive_name": f"framework-layerE{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerB",
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
            {
                "archive_build": 1,
                "install_target": "framework-layerF",
                "archive_name": f"framework-layerF{ARCHIVE_SUFFIX}",
                "required_layers": [
                    "framework-layerE",
                    "framework-layerB",
                    "framework-layerA",
                ],
                "target_platform": BUILD_PLATFORM,
            },
        ],
        "runtimes": [
            {
                "archive_build": 1,
                "install_target": "cpython-3.11",
                "archive_name": f"cpython-3.11{ARCHIVE_SUFFIX}",
                "target_platform": BUILD_PLATFORM,
            },
        ],
    }
}

LastLockedTimes = dict[str, datetime]  # Mapping from install target names to lock times
_CHECKED_KEYS = frozenset(EXPECTED_MANIFEST["layers"]["applications"][0])


def _filter_archive_manifest(archive_manifest: ArchiveBuildMetadata) -> ArchiveSummary:
    """Drop archive manifest fields that aren't covered by this set of test cases."""
    summary: ArchiveSummary = {}
    for key in _CHECKED_KEYS:
        value = archive_manifest.get(key)
        if value is not None:
            summary[key] = value
    return summary


def _filter_manifest(
    manifest: StackPublishingRequest,
) -> tuple[BuildManifest, LastLockedTimes]:
    """Extract manifest fields that are relevant to this set of test cases."""
    filtered_summaries: ArchiveSummaries = {}
    last_locked_times: LastLockedTimes = {}
    for kind, archive_manifests in manifest["layers"].items():
        filtered_summaries[kind] = summaries = []
        for archive_manifest in archive_manifests:
            summaries.append(_filter_archive_manifest(archive_manifest))
            last_locked_times[archive_manifest["install_target"]] = (
                datetime.fromisoformat(archive_manifest["locked_at"])
            )
    return {"layers": filtered_summaries}, last_locked_times


def _tag_manifest(manifest: BuildManifest, expected_tag: str) -> BuildManifest:
    """Add expected build tag to fields that are expected to include the build tag."""
    tagged_summaries: ArchiveSummaries = {}
    for kind, summaries in manifest["layers"].items():
        tagged_summaries[kind] = new_summaries = []
        for summary in summaries:
            new_summary = summary.copy()
            new_summaries.append(new_summary)
            # Archive name has the build tag inserted before the extension
            install_target = summary["install_target"]
            new_summary["archive_name"] = (
                f"{install_target}{expected_tag}{ARCHIVE_SUFFIX}"
            )
    return {"layers": tagged_summaries}


##########################
# Test cases
##########################


class TestMinimalSpec(SpecLoadingTestCase):
    # Test cases that only need the stack specification file

    def test_spec_loading(self) -> None:
        self.check_stack_specification(
            MINIMAL_PROJECT_STACK_SPEC_PATH,
            EXPECTED_ENVIRONMENTS,
            EXPECTED_RUNTIMES,
            EXPECTED_FRAMEWORKS,
            EXPECTED_APPLICATIONS,
        )


class TestMinimalBuildConfig(unittest.TestCase):
    # These test cases don't need the build environment to actually exist

    def setUp(self) -> None:
        # No files are created, so no need to use a temporary directory
        self.stack_spec = StackSpec.load(MINIMAL_PROJECT_STACK_SPEC_PATH)

    def test_default_build_directory(self) -> None:
        stack_spec = self.stack_spec
        build_env = stack_spec.define_build_environment()
        expected_build_path = stack_spec.spec_path.parent
        self.assertEqual(build_env.build_path, expected_build_path)
        # The spec directory necessarily already exists
        self.assertTrue(expected_build_path.exists())

    def test_custom_build_directory_relative(self) -> None:
        stack_spec = self.stack_spec
        build_env = stack_spec.define_build_environment("custom")
        expected_build_path = stack_spec.spec_path.parent / "custom"
        self.assertEqual(build_env.build_path, expected_build_path)
        # Build directory is only created when needed, not immediately
        self.assertFalse(expected_build_path.exists())

    def test_custom_build_directory_user(self) -> None:
        build_env = self.stack_spec.define_build_environment("~/custom")
        expected_build_path = Path.home() / "custom"
        self.assertEqual(build_env.build_path, expected_build_path)
        # Build directory is only created when needed, not immediately
        self.assertFalse(expected_build_path.exists())

    def test_custom_build_directory_absolute(self) -> None:
        expected_build_path = Path("/custom").absolute()  # Add drive info on Windows
        build_env = self.stack_spec.define_build_environment(expected_build_path)
        self.assertEqual(build_env.build_path, expected_build_path)
        # Build directory is only created when needed, not immediately
        self.assertFalse(expected_build_path.exists())

    def test_env_categories_without_lock_files(self) -> None:
        stack_spec = self.stack_spec
        build_env = stack_spec.define_build_environment()
        expected_names = [env.env_name for env in EXPECTED_ENVIRONMENTS]
        all_names = [env.env_name for env in build_env.all_environments()]
        self.assertEqual(all_names, expected_names)
        # No lock files, so all envs should need locking
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = [env.env_name for env in envs_to_lock]
        self.assertEqual(lock_names, expected_names)
        self.assertTrue(all(env.needs_lock() for env in envs_to_lock))
        self.assertTrue(build_env._needs_lock())
        # All envs should be flagged for building
        build_names = [env.env_name for env in build_env.environments_to_build()]
        self.assertEqual(build_names, expected_names)
        # All envs should be flagged for publishing
        publish_names = [env.env_name for env in build_env.environments_to_publish()]
        self.assertEqual(publish_names, expected_names)

    def test_env_categories_with_ops_disabled(self) -> None:
        stack_spec = self.stack_spec
        build_env = stack_spec.define_build_environment()
        # Disable all operations
        # Also check disabling locking entirely overrides the lock reset request
        build_env.select_operations(
            lock=False, build=False, publish=False, reset_lock=True
        )
        expected_names = [env.env_name for env in EXPECTED_ENVIRONMENTS]
        all_envs = list(build_env.all_environments())
        all_names = [env.env_name for env in all_envs]
        self.assertEqual(all_names, expected_names)
        # No envs should be selected for locking
        self.assertEqual(list(build_env.environments_to_lock()), [])
        self.assertFalse(any(env.needs_lock() for env in all_envs))
        self.assertFalse(build_env._needs_lock())
        # No envs should be flagged for building
        self.assertEqual(list(build_env.environments_to_build()), [])
        # No envs should be flagged for publishing
        self.assertEqual(list(build_env.environments_to_publish()), [])


class TestMinimalBuildConfigWithExistingLockFiles(unittest.TestCase):
    # These test cases don't need the build environment to actually exist

    def setUp(self) -> None:
        # Need a temporary directory to avoid cross-test side effects
        working_dir = tempfile.TemporaryDirectory()
        self.addCleanup(working_dir.cleanup)
        self.working_path = working_path = Path(working_dir.name)
        self.build_env = build_env = _define_build_env(working_path)
        self.expected_build_path = working_path / "_build🐸"
        # Mimic the environments already being locked
        build_platform = build_env.build_platform
        lock_dir_path = build_env.requirements_dir_path
        for env in build_env.all_environments():
            env_spec = env.env_spec
            env.env_lock.prepare_lock_inputs()  # Also creates relevant subdir
            requirements_path = env_spec.get_requirements_path(
                build_platform, lock_dir_path
            )
            requirements_path.write_text("")
            env.env_lock.update_lock_metadata()
            self.assertIsNotNone(env.env_lock.load_valid_metadata())
        # Path diffs can get surprisingly long
        self.maxDiff = None

    def check_publishing_request(
        self, publishing_request: StackPublishingRequest
    ) -> None:
        self.assertEqual(_filter_manifest(publishing_request)[0], EXPECTED_MANIFEST)

    def test_default_output_directory(self) -> None:
        build_env = self.build_env
        output_path, publishing_request = build_env.publish_artifacts(dry_run=True)
        # Build folder is used as the default output directory
        expected_output_path = self.expected_build_path
        self.assertEqual(output_path, expected_output_path)
        self.check_publishing_request(publishing_request)
        # The build directory necessarily already exists
        self.assertFalse(expected_output_path.exists())

    def test_custom_output_directory_relative(self) -> None:
        build_env = self.build_env
        output_path, publishing_request = build_env.publish_artifacts(
            "custom", dry_run=True
        )
        expected_output_path = self.working_path / "custom"
        self.assertEqual(output_path, expected_output_path)
        self.check_publishing_request(publishing_request)
        # Dry run doesn't create the output directory
        self.assertFalse(expected_output_path.exists())

    def test_custom_output_directory_user(self) -> None:
        build_env = self.build_env
        output_path, publishing_request = build_env.publish_artifacts(
            "~/custom", dry_run=True
        )
        expected_output_path = Path.home() / "custom"
        self.assertEqual(output_path, expected_output_path)
        self.check_publishing_request(publishing_request)
        # Dry run doesn't create the output directory
        self.assertFalse(expected_output_path.exists())

    def test_custom_output_directory_absolute(self) -> None:
        build_env = self.build_env
        expected_output_path = Path("/custom").absolute()  # Add drive info on Windows
        output_path, publishing_request = build_env.publish_artifacts(
            expected_output_path, dry_run=True
        )
        self.assertEqual(output_path, expected_output_path)
        self.check_publishing_request(publishing_request)
        # Dry run doesn't create the output directory
        self.assertFalse(expected_output_path.exists())

    def test_env_categories_with_lock_files(self) -> None:
        build_env = self.build_env
        expected_names = [env.env_name for env in EXPECTED_ENVIRONMENTS]
        all_names = [env.env_name for env in build_env.all_environments()]
        self.assertEqual(all_names, expected_names)
        # Lock files exist, so no envs should *need* locking,
        # but their lock status should still be checked by default
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = [env.env_name for env in envs_to_lock]
        self.assertEqual(lock_names, expected_names)
        self.assertFalse(any(env.needs_lock() for env in envs_to_lock))
        self.assertFalse(build_env._needs_lock())
        # All envs should be flagged for building
        build_names = [env.env_name for env in build_env.environments_to_build()]
        self.assertEqual(build_names, expected_names)
        # All envs should be flagged for publishing
        publish_names = [env.env_name for env in build_env.environments_to_publish()]
        self.assertEqual(publish_names, expected_names)

    def test_env_categories_lock_with_lock_reset(self) -> None:
        build_env = self.build_env
        # Enable locking with locked requirement resets
        build_env.select_operations(
            lock=True, build=False, publish=False, reset_lock=True
        )
        expected_names = [env.env_name for env in EXPECTED_ENVIRONMENTS]
        all_names = [env.env_name for env in build_env.all_environments()]
        self.assertEqual(all_names, expected_names)
        # Lock reset requested, so all envs should need locking
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = [env.env_name for env in envs_to_lock]
        self.assertEqual(lock_names, expected_names)
        self.assertTrue(all(env.needs_lock() for env in envs_to_lock))
        self.assertTrue(build_env._needs_lock())
        # No envs should be flagged for building
        self.assertEqual(list(build_env.environments_to_build()), [])
        # No envs should be flagged for publishing
        self.assertEqual(list(build_env.environments_to_publish()), [])

    def test_env_categories_lock_layers_with_lock_reset(self) -> None:
        build_env = self.build_env
        # Enable locking with locked requirement resets
        all_names = [env.env_name for env in build_env.all_environments()]
        build_env.select_layers(
            lock=True,
            build=False,
            publish=False,
            include=all_names,
            reset_locks=all_names,
        )
        # Lock reset requested, so all envs should need locking
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = [env.env_name for env in envs_to_lock]
        self.assertEqual(lock_names, all_names)
        self.assertTrue(all(env.needs_lock() for env in envs_to_lock))
        self.assertTrue(build_env._needs_lock())
        # No envs should be flagged for building
        self.assertEqual(list(build_env.environments_to_build()), [])
        # No envs should be flagged for publishing
        self.assertEqual(list(build_env.environments_to_publish()), [])

    def test_env_categories_selective_lock_with_lock_reset(self) -> None:
        build_env = self.build_env
        # Locking only "framework-layerA" with the default settings:
        # * should omit "app-no-framework" entirely
        # * should only lock the runtime layer if necessary (no implicit reset)
        # * should lock and reset everything else directly or as a derived layer
        locked_layer = "framework-layerA"
        no_lock = {EnvNameBuild("app-no-framework")}
        no_reset = build_env.filter_layers(["cpython-*"])[0] | no_lock
        assert len(no_reset) > 1

        all_names = {env.env_name for env in build_env.all_environments()}
        included, _ = build_env.filter_layers([locked_layer])
        assert included == {"framework-layerA"}
        build_env.select_layers(
            lock=True,
            build=False,
            publish=False,
            include=included,
            reset_locks=all_names,
        )
        # Ensure expected envs want and need locking
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = {env.env_name for env in envs_to_lock}
        self.assertEqual(lock_names, all_names - no_lock)
        self.assertTrue(
            all(
                env.needs_lock() for env in envs_to_lock if env.env_name not in no_reset
            )
        )
        self.assertTrue(build_env._needs_lock())
        # Mark the runtime environment as unlocked
        for rt_env in build_env.runtimes.values():
            rt_env.env_lock._purge_lock()
        # Check the runtime environment has been added to layers to lock
        envs_to_lock = list(build_env.environments_to_lock())
        lock_names = {env.env_name for env in envs_to_lock}
        self.assertEqual(lock_names, all_names - no_lock)
        self.assertTrue(all(env.needs_lock() for env in envs_to_lock))
        self.assertTrue(build_env._needs_lock())


class TestMinimalBuild(DeploymentTestCase):
    # Test cases that actually create the build environment folders

    working_path: Path
    build_env: BuildEnvironment

    def setUp(self) -> None:
        # Need a temporary directory to avoid cross-test side effects
        working_dir = tempfile.TemporaryDirectory()
        self.addCleanup(working_dir.cleanup)
        self.working_path = working_path = Path(working_dir.name)
        self.build_env = _define_build_env(working_path)

    def assertRecentlyLocked(
        self, last_locked_times: LastLockedTimes, minimum_lock_time: datetime
    ) -> None:
        for install_target, last_locked in last_locked_times.items():
            # Use a tuple comparison so the install_target value gets
            # reported without needing to define nested subtests
            self.assertGreaterEqual(
                (install_target, last_locked), (install_target, minimum_lock_time)
            )

    @staticmethod
    def _load_archive_summary(metadata_path: Path) -> ArchiveSummary:
        with metadata_path.open("r", encoding="utf-8") as f:
            return _filter_archive_manifest(json.load(f))

    @staticmethod
    def _load_build_manifest(metadata_path: Path) -> BuildManifest:
        with metadata_path.open("r", encoding="utf-8") as f:
            return _filter_manifest(json.load(f))[0]

    def mock_index_config_options(
        self, reference_config: PackageIndexConfig | None = None
    ) -> None:
        # Mock the index configs in order to check for
        # expected CLI argument lookups
        for env in self.build_env.all_environments():
            if reference_config is None:
                env_reference_config = env.index_config
            else:
                env_reference_config = reference_config
            env.index_config = make_mock_index_config(env_reference_config)

    def check_publication_result(
        self,
        publication_result: PublishedArchivePaths,
        dry_run_result: BuildManifest,
        expected_tag: str | None,
    ) -> None:
        # Build dir is used as the default output path
        expected_output_path = self.build_env.build_path
        expected_metadata_path = expected_output_path / BuildEnvironment.METADATA_DIR
        expected_env_metadata_path = (
            expected_metadata_path / BuildEnvironment.METADATA_ENV_DIR
        )
        if expected_tag is None:
            expected_metadata_name = "venvstacks.json"
            expected_snippet_suffix = ".json"
        else:
            expected_metadata_name = f"venvstacks{expected_tag}.json"
            expected_snippet_suffix = f"{expected_tag}.json"
        manifest_path, snippet_paths, archive_paths = publication_result
        # Check overall manifest file
        expected_manifest_path = expected_metadata_path / expected_metadata_name
        self.assertEqual(manifest_path, expected_manifest_path)
        manifest_data = self._load_build_manifest(manifest_path)
        self.assertEqual(manifest_data, dry_run_result)
        # Check individual archive manifests
        expected_summaries: dict[str, ArchiveSummary] = {}
        for archive_summaries in dry_run_result["layers"].values():
            for archive_summary in archive_summaries:
                install_target = archive_summary["install_target"]
                expected_summaries[install_target] = archive_summary
        for snippet_path in snippet_paths:
            archive_summary = self._load_archive_summary(snippet_path)
            install_target = archive_summary["install_target"]
            expected_snippet_name = f"{install_target}{expected_snippet_suffix}"
            expected_snippet_path = expected_env_metadata_path / expected_snippet_name
            self.assertEqual(snippet_path, expected_snippet_path)
            self.assertEqual(archive_summary, expected_summaries[install_target])
        # Check the names and location of the generated archives
        expected_archive_paths: list[Path] = []
        for archive_summaries in dry_run_result["layers"].values():
            for archive_summary in archive_summaries:
                expected_archive_path = (
                    expected_output_path / archive_summary["archive_name"]
                )
                expected_archive_paths.append(expected_archive_path)
        expected_archive_paths.sort()
        self.assertEqual(sorted(archive_paths), expected_archive_paths)

    @staticmethod
    def _run_postinstall(env_path: Path) -> None:
        postinstall_script = env_path / "postinstall.py"
        if postinstall_script.exists():
            # Post-installation scripts are required to work even when they're
            # executed with an entirely unrelated Python installation
            capture_python_output(
                [sys.executable, "-X", "utf8", "-I", str(postinstall_script)]
            )

    def check_archive_deployment(self, published_paths: PublishedArchivePaths) -> None:
        metadata_path, snippet_paths, archive_paths = published_paths
        published_manifests = ManifestData(metadata_path, snippet_paths)
        # TODO: read the base Python path for each environment from the metadata
        #       https://github.com/lmstudio-ai/venvstacks/issues/19
        # TODO: figure out a more robust way of handling Windows potentially still
        #       having the Python executables in the environment open when the
        #       parent process tries to clean up the deployment directory.
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as deployment_dir:
            # Extract archives
            deployment_path = Path(deployment_dir)
            env_name_to_path: dict[EnvNameDeploy, Path] = {}
            expected_dirs: list[str] = []
            for archive_metadata, archive_path in zip(
                published_manifests.snippet_data, archive_paths
            ):
                if ".tar" in archive_path.suffixes:
                    # Layered env tar archives typically have symlinks to their runtime environment
                    shutil.unpack_archive(
                        archive_path, deployment_path, filter="fully_trusted"
                    )
                else:
                    shutil.unpack_archive(archive_path, deployment_path)
                env_name = EnvNameDeploy(archive_metadata["install_target"])
                self.assertEqual(archive_path.name[: len(env_name)], env_name)
                expected_dirs.append(env_name)
                env_path = deployment_path / env_name
                env_name_to_path[env_name] = env_path
            self.assertCountEqual(
                [p.name for p in deployment_path.iterdir()], expected_dirs
            )
            # Run the post install scripts
            self.assertTrue(published_manifests.combined_data)
            layered_metadata = published_manifests.combined_data["layers"]
            base_runtime_env_name = layered_metadata["runtimes"][0]["install_target"]
            env_path = env_name_to_path[base_runtime_env_name]
            self._run_postinstall(env_path)
            for env_name, env_path in env_name_to_path.items():
                if env_name == base_runtime_env_name:
                    # Already configured
                    continue
                self._run_postinstall(env_path)

            def get_deployed_env_details(
                env: ArchiveMetadata,
            ) -> tuple[EnvNameDeploy, Path, list[str]]:
                env_name = env["install_target"]
                env_path = env_name_to_path[env_name]
                env_python = get_env_python(env_path)
                env_sys_path = get_sys_path(env_python)
                return env_name, env_path, env_sys_path

            self.check_deployed_environments(layered_metadata, get_deployed_env_details)

    def test_create_environments(self) -> None:
        # Faster test to check the links between build envs are set up correctly
        # (if this fails, there's no point even trying the full slow test case)
        build_env = self.build_env
        build_env.create_environments()
        self.check_build_environments(self.build_env.all_environments())

    @pytest.mark.slow
    def test_locking_and_publishing(self) -> None:
        # This is organised as subtests in a monolithic test sequence to reduce CI overhead
        # Separating the tests wouldn't really make them independent, unless the outputs of
        # the earlier steps were checked in for use when testing the later steps.
        # Actually configuring and building the environments is executed outside the subtest
        # declarations, since actual build failures need to fail the entire test method.
        subtests_started = subtests_passed = 0  # Track subtest failures
        build_env = self.build_env
        self.mock_index_config_options()
        platform_tag = build_env.build_platform
        expected_tag = f"-{platform_tag}"
        versioned_tag = (
            f"{expected_tag}-1"  # No previous metadata when running the test
        )
        expected_dry_run_result = EXPECTED_MANIFEST
        expected_tagged_dry_run_result = _tag_manifest(EXPECTED_MANIFEST, versioned_tag)
        minimum_lock_time = datetime.now(timezone.utc)
        # Ensure the locking and publication steps always run for all environments
        build_env.select_operations(lock=True, build=True, publish=True)
        self.assertTrue(build_env._needs_lock())
        # Handle running this test case repeatedly in a local checkout
        for env in build_env.all_environments():
            env.env_lock._purge_lock()
        # Create and link the layer build environments
        build_env.create_environments()
        # Don't even try to continue if the environments aren't properly linked
        self.check_build_environments(self.build_env.all_environments())
        # Test stage: check dry run metadata results are as expected
        subtests_started += 1
        with self.subTest("Check untagged dry run"):
            dry_run_result, dry_run_last_locked_times = _filter_manifest(
                build_env.publish_artifacts(dry_run=True)[1]
            )
            self.assertEqual(dry_run_result, expected_dry_run_result)
            self.assertRecentlyLocked(dry_run_last_locked_times, minimum_lock_time)
            # Check for expected subprocess argument lookups
            for env in self.build_env.all_environments():
                # First environment build: lock with uv, install with pip
                mock_compile = cast(Mock, env.index_config._get_uv_pip_compile_args)
                mock_compile.assert_called_once_with()
                mock_compile.reset_mock()
                mock_install = cast(Mock, env.index_config._get_pip_install_args)
                mock_install.assert_called_once_with()
                mock_install.reset_mock()
            subtests_passed += 1
        subtests_started += 1
        with self.subTest("Check tagged dry run"):
            tagged_dry_run_result, tagged_last_locked_times = _filter_manifest(
                build_env.publish_artifacts(dry_run=True, tag_outputs=True)[1]
            )
            self.assertEqual(tagged_dry_run_result, expected_tagged_dry_run_result)
            self.assertEqual(tagged_last_locked_times, dry_run_last_locked_times)
            subtests_passed += 1
        # Test stage: ensure lock timestamps don't change when requirements don't change
        build_env.lock_environments()
        self.assertFalse(build_env._needs_lock())
        subtests_started += 1
        with self.subTest("Check lock timestamps don't change for stable requirements"):
            stable_dry_run_result, stable_last_locked_times = _filter_manifest(
                build_env.publish_artifacts(dry_run=True)[1]
            )
            self.assertEqual(stable_dry_run_result, expected_dry_run_result)
            self.assertEqual(stable_last_locked_times, dry_run_last_locked_times)
            # Check for expected subprocess argument lookups
            for env in self.build_env.all_environments():
                # The lock file is recreated, the timestamp metadata just doesn't
                # get updated if the hash of the contents doesn't change
                mock_compile = cast(Mock, env.index_config._get_uv_pip_compile_args)
                mock_compile.assert_called_once_with()
                mock_compile.reset_mock()
                mock_install = cast(Mock, env.index_config._get_pip_install_args)
                mock_install.assert_not_called()
            subtests_passed += 1
        # Test stage: ensure lock timestamps *do* change when the requirements "change"
        for env in build_env.all_environments():
            # Rather than actually make the hash change, instead change the hash *records*
            env_lock = env.env_lock
            env_lock._requirements_hash = "ensure requirements appear to have changed"
            env_lock._write_lock_metadata()
        minimum_relock_time = datetime.now(timezone.utc)
        build_env.lock_environments()
        subtests_started += 1
        with self.subTest("Check lock timestamps change for updated requirements"):
            relocked_dry_run_result, relocked_last_locked_times = _filter_manifest(
                build_env.publish_artifacts(dry_run=True)[1]
            )
            self.assertEqual(relocked_dry_run_result, expected_dry_run_result)
            self.assertGreater(minimum_relock_time, minimum_lock_time)
            self.assertRecentlyLocked(relocked_last_locked_times, minimum_relock_time)
            # Check for expected subprocess argument lookups
            for env in self.build_env.all_environments():
                # Locked, but not rebuilt, so only uv should be called
                mock_compile = cast(Mock, env.index_config._get_uv_pip_compile_args)
                mock_compile.assert_called_once_with()
                mock_compile.reset_mock()
                mock_install = cast(Mock, env.index_config._get_pip_install_args)
                mock_install.assert_not_called()
            subtests_passed += 1
        # Test stage: ensure exported environments allow launch module execution
        subtests_started += 1
        with self.subTest("Check environment export"):
            export_path = self.working_path / "_export🦎"
            export_result = build_env.export_environments(export_path)
            self.check_environment_exports(export_path, export_result)
            subtests_passed += 1
        # Test stage: ensure published archives and manifests have the expected name
        #             and that unpacking them allows launch module execution
        subtests_started += 1
        with self.subTest("Check untagged publication"):
            publication_result = build_env.publish_artifacts()
            self.check_publication_result(
                publication_result, dry_run_result, expected_tag=None
            )
            self.check_archive_deployment(publication_result)
            subtests_passed += 1
        subtests_started += 1
        with self.subTest("Check tagged publication"):
            tagged_publication_result = build_env.publish_artifacts(tag_outputs=True)
            self.check_publication_result(
                tagged_publication_result, tagged_dry_run_result, expected_tag
            )
            self.check_archive_deployment(tagged_publication_result)
            subtests_passed += 1
        # TODO: Add another test stage that confirms build versions increment as expected

        # Work around pytest-subtests not failing the test case when subtests fail
        # https://github.com/pytest-dev/pytest-subtests/issues/76
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )



"""Basic tests for venvstacks package components."""

from importlib.metadata import version as pkg_version

import pytest

from support import get_pinned_dev_packages


def test_python_api_import() -> None:
    from venvstacks import stacks

    assert hasattr(stacks, "StackSpec")


PINNED_DEV_PACKAGES = sorted(get_pinned_dev_packages().items())


@pytest.mark.parametrize("pkg_name,version", PINNED_DEV_PACKAGES)
def test_pinned_dev_packages(pkg_name: str, version: str) -> None:
    assert pkg_version(pkg_name) == version


# TODO: The assorted utility classes and functions added to stacks.py when it was
#       a mostly standalone script should be separated out and unit tested



"""Test loading assorted stack specifications."""

from pathlib import Path

import pytest

from venvstacks.stacks import LayerSpecError, StackSpec

##################################
# Stack spec loading test helpers
##################################


_THIS_PATH = Path(__file__)
TEST_SPEC_PATH = _THIS_PATH.parent / "stack_specs"


def _load_stack_spec(spec_name: str) -> StackSpec:
    """Load the named stack specification."""
    spec_path = TEST_SPEC_PATH / spec_name
    return StackSpec.load(spec_path)


##########################
# Test cases
##########################


def test_at_symbol_in_layer_names() -> None:
    stack_spec = _load_stack_spec("at_symbol.toml")
    runtimes = list(stack_spec.all_environment_specs())
    assert len(runtimes) == 2
    unversioned, versioned = runtimes
    # Check the unversioned layer
    assert unversioned.name == "cpython@3.11"
    assert not unversioned.versioned
    # Check the versioned layer
    assert versioned.name == "cpython@3.12"
    assert versioned.versioned


def test_future_warning_for_fully_versioned_name() -> None:
    expected_msg = (
        "Converting legacy.*'fully_versioned_name'.*'python_implementation'.*'runtime'"
    )
    with pytest.warns(FutureWarning, match=expected_msg):
        stack_spec = _load_stack_spec("warning_fully_versioned.toml")
    runtimes = list(stack_spec.all_environment_specs())
    assert len(runtimes) == 1
    (runtime,) = runtimes


def test_future_warning_for_build_requirements() -> None:
    # This actually emits the warning 3 times, but we don't check for that
    # (the fact the spec loads indicates the field is dropped for all layers)
    expected_msg = "Dropping legacy.*'build_requirements'.*'(runtime|fw|app)'"
    with pytest.warns(FutureWarning, match=expected_msg):
        stack_spec = _load_stack_spec("warning_build_requirements.toml")
    layers = list(stack_spec.all_environment_specs())
    assert len(layers) == 3
    for layer in layers:
        assert not hasattr(layer, "build_requirements")


EXPECTED_ERRORS = {
    "error_inconsistent_runtimes.toml": (LayerSpecError, "inconsistent frameworks"),
    "error_layer_dep_C3_conflict.toml": (
        LayerSpecError,
        "linearization failed.*['layerC', 'layerD'].*['layerD', 'layerC']",
    ),
    "error_layer_dep_cycle.toml": (LayerSpecError, "unknown framework"),
    "error_layer_dep_forward_reference.toml": (LayerSpecError, "unknown framework"),
    "error_missing_launch_module.toml": (
        LayerSpecError,
        "launch module.*does not exist",
    ),
    "error_unknown_framework.toml": (LayerSpecError, "unknown framework"),
    "error_unknown_runtime.toml": (LayerSpecError, "unknown runtime"),
}


def test_error_case_results_are_defined() -> None:
    # Ensure any new error cases that are added have expected errors defined
    defined_error_cases = sorted(p.name for p in TEST_SPEC_PATH.glob("error_*"))
    assert defined_error_cases == sorted(EXPECTED_ERRORS)


@pytest.mark.parametrize("spec_path", EXPECTED_ERRORS)
def test_stack_spec_error_case(spec_path: str) -> None:
    expected_exc, expected_match = EXPECTED_ERRORS[spec_path]
    with pytest.raises(expected_exc, match=expected_match):
        _load_stack_spec(spec_path)



"""Test cases for CLI invocation."""

import subprocess
import sys

from collections import namedtuple
from contextlib import contextmanager
from dataclasses import dataclass, field
from fnmatch import fnmatch
from pathlib import Path
from traceback import format_exception
from types import ModuleType
from typing import (
    Any,
    cast,
    Generator,
    get_type_hints,
    Iterable,
    Iterator,
    Self,
    Sequence,
    Set,
)
from unittest.mock import call as mock_call, create_autospec, MagicMock, patch

import pytest

import click.testing
import typer
from typer.models import ArgumentInfo, OptionInfo
from typer.testing import CliRunner

from venvstacks import cli
from venvstacks.stacks import (
    BuildEnvironment,
    EnvNameBuild,
    EnvironmentLock,
    PackageIndexConfig,
)
from venvstacks._util import run_python_command_unchecked

from support import requires_venv


def report_traceback(exc: BaseException | None) -> str:
    if exc is None:
        return "Expected exception was not raised"
    return "\n".join(format_exception(exc))


def _mock_path(contents: str | None = None) -> Any:
    mock_path = create_autospec(Path, spec_set=True, instance=True)
    mock_path.exists.return_value = True
    if contents is None:
        # Pretend this is a directory path
        mock_path.is_dir.return_value = True
    else:
        # Pretend this is a readable file path
        mock_path.read_text.return_value = contents
        mock_path.is_dir.return_value = False
    return mock_path


def mock_environment_locking(*, clean: bool = False) -> Sequence[EnvironmentLock]:
    return []


def mock_output_generation(output_dir: Path, *, dry_run: bool = False, **_: Any) -> Any:
    if dry_run:
        return _mock_path(), {}
    return _mock_path("{}"), [], []


MOCK_LAYERS = ["runtime", "framework-A", "framework-B", "app-1", "app-2", "app-3"]

MockEnv = namedtuple("MockEnv", ("env_name",))


def mock_environment_iterator() -> Iterator[MockEnv]:
    for env_name in MOCK_LAYERS:
        yield MockEnv(env_name)


def mock_layer_filtering(patterns: Iterable[str]) -> tuple[Set[EnvNameBuild], Set[str]]:
    matching = set(
        env for env in MOCK_LAYERS if any(fnmatch(env, pattern) for pattern in patterns)
    )
    return cast(Set[EnvNameBuild], matching), set()


@dataclass(repr=False, eq=False)
class MockedRunner:
    app: typer.Typer
    mocked_stack_spec: MagicMock

    runner: CliRunner = field(init=False)
    mocked_build_env: MagicMock = field(init=False)

    def __post_init__(self) -> None:
        # Note: `mock` doesn't quite handle `dataclass` instances correctly
        #       (see https://github.com/python/cpython/issues/124176 for details)
        #       However, the CLI doesn't currently try to access any of the missing
        #       attributes, so the autospec mocking here is sufficient in practice.
        self.runner = CliRunner()
        mocked_stack_spec = self.mocked_stack_spec
        # Use the patched in mock as the sole defined spec instance
        mocked_stack_spec.load.return_value = mocked_stack_spec
        # Patch in a mocked build environment
        mocked_build_env = create_autospec(
            BuildEnvironment, spec_set=True, instance=True
        )
        # use the mocked build environment in test cases
        mocked_stack_spec.define_build_environment.return_value = mocked_build_env
        self.mocked_build_env = mocked_build_env
        # Control the result of the environment locking step
        mocked_build_env.lock_environments.side_effect = mock_environment_locking
        # Control the result of the layer filter step
        mocked_build_env.all_environments.side_effect = mock_environment_iterator
        mocked_build_env.filter_layers.side_effect = mock_layer_filtering
        # Control the result of artifact publication and environment exports
        mocked_build_env.publish_artifacts.side_effect = mock_output_generation
        mocked_build_env.export_environments.side_effect = mock_output_generation

    @classmethod
    @contextmanager
    def cli_patch_installed(cls, cli_module: ModuleType) -> Iterator[Self]:
        """Patch the given CLI module to invoke a mocked StackSpec instance."""
        app = cli_module._cli
        patch_cm = patch.object(cli_module, "StackSpec", autospec=True, spec_set=True)
        with patch_cm as mocked_stack_spec:
            yield cls(app, mocked_stack_spec)

    def invoke(self, cli_args: list[str]) -> click.testing.Result:
        return self.runner.invoke(self.app, cli_args)

    def assert_build_config(
        self, expected_build_dir: str, expected_index_config: PackageIndexConfig
    ) -> None:
        """Check environment build path and index configuration details."""
        env_definition = self.mocked_stack_spec.define_build_environment
        env_definition.assert_called_with(expected_build_dir, expected_index_config)

    _OUTPUT_METHODS = {
        "build": "publish_artifacts",
        "publish": "publish_artifacts",
        "local-export": "export_environments",
    }

    def get_output_method(self, command: str) -> MagicMock:
        """Return the Mock expected to be called for the given output command."""
        output_method_name = self._OUTPUT_METHODS[command]
        return cast(MagicMock, getattr(self.mocked_build_env, output_method_name))

    _DEFAULT_OUTPUT_DIRS = {
        "build": "_artifacts",
        "publish": "_artifacts",
        "local-export": "_export",
    }

    def get_default_output_dir(self, command: str) -> str:
        """Return the Mock expected to be called for the given output command."""
        return self._DEFAULT_OUTPUT_DIRS[command]


@pytest.fixture
def mocked_runner() -> Generator[MockedRunner, None, None]:
    with MockedRunner.cli_patch_installed(cli) as mocked_app:
        yield mocked_app


class TestTopLevelCommand:
    def test_implicit_help(self, mocked_runner: MockedRunner) -> None:
        result = mocked_runner.invoke([])
        # Usage message should suggest indirect execution
        assert "Usage: python -m venvstacks [" in result.stdout
        # Top-level callback docstring is used as the overall CLI help text
        cli_help = cli.handle_app_options.__doc__
        assert cli_help is not None
        assert cli_help.strip() in result.stdout
        # Subcommands are listed in the top level help
        assert str(cli.build.__name__) in result.stdout
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @requires_venv("Entry point test")
    def test_entry_point_help_raw(self) -> None:
        expected_entry_point = Path(sys.executable).parent / "venvstacks"
        command = [str(expected_entry_point), "--help"]
        result = run_python_command_unchecked(
            command, text=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if result.stdout is not None:
            # Usage message should suggest direct execution
            assert b"Usage: venvstacks [" in result.stdout
            # Top-level callback docstring is used as the overall CLI help text
            cli_help = cli.handle_app_options.__doc__
            assert cli_help is not None
            assert cli_help.strip().encode() in result.stdout
        # Check operation result last to ensure test results are as informative as possible
        assert result.returncode == 0
        assert result.stdout is not None

    @requires_venv("Entry point test")
    def test_entry_point_help(self) -> None:
        expected_entry_point = Path(sys.executable).parent / "venvstacks"
        command = [str(expected_entry_point), "--help"]
        result = run_python_command_unchecked(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if result.stdout is not None:
            # Usage message should suggest direct execution
            assert "Usage: venvstacks [" in result.stdout
            # Top-level callback docstring is used as the overall CLI help text
            cli_help = cli.handle_app_options.__doc__
            assert cli_help is not None
            assert cli_help.strip() in result.stdout
        # Check operation result last to ensure test results are as informative as possible
        assert result.returncode == 0
        assert result.stdout is not None

    def test_module_execution(self) -> None:
        # TODO: `coverage.py` isn't picking this up as executing `venvstacks/__main__.py`
        #       (even an indirect invocation via the runpy module doesn't get detected)
        command = [sys.executable, "-m", "venvstacks", "--help"]
        result = run_python_command_unchecked(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if result.stdout is not None:
            # Usage message should suggest indirect execution
            assert "Usage: python -m venvstacks [" in result.stdout
            # Top-level callback docstring is used as the overall CLI help text
            cli_help = cli.handle_app_options.__doc__
            assert cli_help is not None
            assert cli_help.strip() in result.stdout
        # Check operation result last to ensure test results are as informative as possible
        assert result.returncode == 0
        assert result.stdout is not None


EXPECTED_USAGE_PREFIX = "Usage: python -m venvstacks "
EXPECTED_SUBCOMMANDS = ["lock", "build", "local-export", "publish"]
NO_SPEC_PATH: list[str] = []
NEEDS_SPEC_PATH = sorted(set(EXPECTED_SUBCOMMANDS) - set(NO_SPEC_PATH))
ACCEPTS_BUILD_DIR = ["lock", "build", "local-export", "publish"]
ACCEPTS_OUTPUT_DIR = ["build", "local-export", "publish"]
ACCEPTS_INDEX_CONFIG = ["lock", "build"]
ACCEPTS_RESET_LOCK = ["lock", "build"]


def _get_default_index_config(command: str) -> PackageIndexConfig:
    if command in ACCEPTS_INDEX_CONFIG:
        return PackageIndexConfig()
    # Commands that don't support index access should turn it off in their config
    return PackageIndexConfig.disabled()


ARGUMENT_PREFIX = "_CLI_ARG"
OPTION_PREFIXES = {
    bool: "_CLI_OPT_FLAG",
    bool | None: "_CLI_OPT_TRISTATE",
    list[str] | None: "_CLI_OPT_STRLIST",
    str: "_CLI_OPT_STR",
}


class TestSubcommands:
    @pytest.mark.parametrize("command", EXPECTED_SUBCOMMANDS)
    def test_internal_consistency(self, command: str) -> None:
        # Check all CLI annotations are internally consistent:
        #   * ensures all used _CLI prefixes are consistent with their types
        #   * ensures all arg names are consistent with annotation suffixes
        command_impl_name = command.replace("-", "_")
        command_impl = getattr(cli, command_impl_name)
        annotations = get_type_hints(command_impl, include_extras=True)
        for arg_name, arg_annotation in annotations.items():
            if arg_name == "return":
                assert arg_annotation is type(None)
                continue
            arg_kind = type(arg_annotation.__metadata__[0])
            assert arg_kind is ArgumentInfo or arg_kind is OptionInfo
            arg_type = arg_annotation.__origin__
            if arg_kind is ArgumentInfo:
                expected_text_prefix = ARGUMENT_PREFIX
            else:
                option_prefix = OPTION_PREFIXES.get(arg_type)
                if option_prefix is None:
                    assert False, f"No CLI option prefix defined for {arg_type!r}"
                expected_text_prefix = option_prefix
            expected_annotation_name = f"{expected_text_prefix}_{arg_name}"
            named_annotation = getattr(cli, expected_annotation_name)
            assert named_annotation == arg_annotation

    @pytest.mark.parametrize("command", EXPECTED_SUBCOMMANDS)
    def test_help_option(self, mocked_runner: MockedRunner, command: str) -> None:
        result = mocked_runner.invoke([command, "--help"])
        # Command implementation docstring is used as the subcommand help text
        command_impl_name = command.replace("-", "_")
        command_impl = getattr(cli, command_impl_name)
        cli_help = command_impl.__doc__
        assert cli_help is not None
        assert cli_help.strip() in result.stdout
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        mocked_stack_spec.define_build_environment.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @pytest.mark.parametrize("command", NEEDS_SPEC_PATH)
    def test_usage_error(self, mocked_runner: MockedRunner, command: str) -> None:
        result = mocked_runner.invoke([command])
        # No overall help in a usage error
        command_impl_name = command.replace("-", "_")
        command_impl = getattr(cli, command_impl_name)
        cli_help = command_impl.__doc__
        assert cli_help is not None
        assert cli_help.strip() not in result.stdout
        # Should complain about the missing required argument
        assert result.stdout[: len(EXPECTED_USAGE_PREFIX)] == EXPECTED_USAGE_PREFIX
        assert "Missing argument 'SPEC_PATH'" in result.stdout
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        mocked_stack_spec.define_build_environment.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert isinstance(result.exception, SystemExit), report_traceback(
            result.exception
        )
        assert result.exit_code == 2

    @pytest.mark.parametrize("command", ACCEPTS_BUILD_DIR)
    def test_build_dir_configuration(
        self, mocked_runner: MockedRunner, command: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(
            [command, "--build-dir", "custom", spec_path_to_mock]
        )
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "custom"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        expected_index_config = _get_default_index_config(command)
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        if command in ACCEPTS_OUTPUT_DIR:
            # Only check the output path (other tests check the other parameters)
            output_method = mocked_runner.get_output_method(command)
            expected_output_dir = mocked_runner.get_default_output_dir(command)
            output_method.assert_called_once()
            assert output_method.call_args.args == (expected_output_dir,)
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @pytest.mark.parametrize("command", ACCEPTS_OUTPUT_DIR)
    def test_output_dir_configuration(
        self, mocked_runner: MockedRunner, command: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(
            [command, "--output-dir", "custom", spec_path_to_mock]
        )
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        expected_output_dir = "custom"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        expected_index_config = _get_default_index_config(command)
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        # Only check the output path (other tests check the other parameters)
        output_method = mocked_runner.get_output_method(command)
        output_method.assert_called_once()
        assert output_method.call_args.args == (expected_output_dir,)
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    # All commands that accept --build-dir should also accept --include
    @pytest.mark.parametrize("command", ACCEPTS_BUILD_DIR)
    def test_layer_selection(self, mocked_runner: MockedRunner, command: str) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(
            [command, "--include", "app-*", spec_path_to_mock]
        )
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        expected_index_config = _get_default_index_config(command)
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        # Only check the layer selection invocation
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_not_called()
        mocked_build_env.filter_layers.assert_called_once_with(
            ["app-*"],
        )
        layer_selection = mocked_build_env.select_layers
        layer_selection.assert_called_once()
        expected_layers = set(env for env in MOCK_LAYERS if env.startswith("app-"))
        assert layer_selection.call_args.args == (expected_layers,)
        assert layer_selection.call_args.kwargs["reset_locks"] == ()
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @pytest.mark.parametrize("command", ACCEPTS_RESET_LOCK)
    def test_lock_reset_selection(
        self, mocked_runner: MockedRunner, command: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(
            [command, "--reset-lock", "app-*", spec_path_to_mock]
        )
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        expected_index_config = _get_default_index_config(command)
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        # Only check the layer selection invocation
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_not_called()
        mocked_build_env.filter_layers.assert_called_once_with(
            ["app-*"],
        )
        layer_selection = mocked_build_env.select_layers
        layer_selection.assert_called_once()
        assert layer_selection.call_args.args == (MOCK_LAYERS,)
        expected_layers = set(env for env in MOCK_LAYERS if env.startswith("app-"))
        assert layer_selection.call_args.kwargs["reset_locks"] == expected_layers
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @pytest.mark.parametrize("command", ACCEPTS_RESET_LOCK)
    def test_layer_selection_with_lock_reset(
        self, mocked_runner: MockedRunner, command: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(
            [command, "--include", "app-*", "--reset-lock", "*", spec_path_to_mock]
        )
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        expected_index_config = _get_default_index_config(command)
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        # Only check the layer selection invocation
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_not_called()
        mocked_build_env.filter_layers.assert_has_calls(
            (
                mock_call(["app-*"]),
                mock_call(["*"]),
            )
        )
        layer_selection = mocked_build_env.select_layers
        layer_selection.assert_called_once()
        expected_layers = set(env for env in MOCK_LAYERS if env.startswith("app-"))
        assert layer_selection.call_args.args == (expected_layers,)
        assert layer_selection.call_args.kwargs["reset_locks"] == set(MOCK_LAYERS)
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @staticmethod
    def _cli_args_case_id(cli_args_test_case: tuple[Any, ...]) -> str:
        extra_cli_args: tuple[str, ...] = cli_args_test_case[0]
        return f"({' '.join(extra_cli_args)})"

    # CLI option handling test cases for package index access configuration
    IndexConfigCase = tuple[tuple[str, ...], PackageIndexConfig]
    _INDEX_CONFIG_ARGS = (
        # Define how the relevant CLI options map to build environment config settings
        (
            (),
            PackageIndexConfig(
                query_default_index=True,
                local_wheel_dirs=None,
            ),
        ),
        (
            ("--no-index",),
            PackageIndexConfig(
                query_default_index=False,
                local_wheel_dirs=None,
            ),
        ),
        (
            ("--local-wheels", "/some_dir", "--local-wheels", "some/other/dir"),
            PackageIndexConfig(
                query_default_index=True,
                local_wheel_dirs=["/some_dir", "some/other/dir"],
            ),
        ),
    )

    @pytest.mark.parametrize("cli_test_case", _INDEX_CONFIG_ARGS, ids=_cli_args_case_id)
    @pytest.mark.parametrize("command", ACCEPTS_INDEX_CONFIG)
    def test_index_options(
        self, mocked_runner: MockedRunner, command: str, cli_test_case: IndexConfigCase
    ) -> None:
        extra_cli_args, expected_index_config = cli_test_case
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke([command, *extra_cli_args, spec_path_to_mock])
        # Always loads the stack spec
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        # Check build environment is created with the expected options
        mocked_runner.assert_build_config(expected_build_dir, expected_index_config)
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    # Specific CLI option handling test cases for the "build" subcommand
    BuildFlagCase = tuple[
        tuple[str, ...], dict[str, bool], dict[str, bool], dict[str, bool]
    ]
    _BUILD_OPTIONS = (
        # Define how the various flags in the CLI build subcommand map to API method options
        # CLI vs API distinction: `--[no-]publish` controls the `dry_run` publication flag
        #                         rather than controlling the `publish` operation selection
        # `--include` and its related options are tested separately
        (
            (),
            dict(lock=False, build=True, publish=True),  # select_operations
            dict(clean=False, lock=False),  # create_environments
            dict(dry_run=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--lock",),
            dict(lock=True, build=True, publish=True),  # select_operations
            dict(clean=False, lock=True),  # create_environments
            dict(dry_run=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--publish",),
            dict(lock=False, build=True, publish=True),  # select_operations
            dict(clean=False, lock=False),  # create_environments
            dict(force=False, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--clean",),
            dict(lock=False, build=True, publish=True),  # select_operations
            dict(clean=True, lock=False),  # create_environments
            dict(dry_run=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            (
                "--publish",
                "--clean",
            ),
            dict(lock=False, build=True, publish=True),  # select_operations
            dict(clean=True, lock=False),  # create_environments
            dict(force=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--tag-outputs",),
            dict(lock=False, build=True, publish=True),  # select_operations
            dict(clean=False, lock=False),  # create_environments
            dict(dry_run=True, tag_outputs=True),  # publish_artifacts
        ),
        (
            ("--lock", "--build", "--publish", "--clean", "--tag-outputs"),
            dict(lock=True, build=True, publish=True),  # select_operations
            dict(clean=True, lock=True),  # create_environments
            dict(force=True, tag_outputs=True),  # publish_artifacts
        ),
        (
            (
                "--no-lock",
                "--no-build",
                "--no-publish",
                "--no-clean",
                "--no-tag-outputs",
            ),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(clean=False, lock=False),  # create_environments
            dict(dry_run=True, tag_outputs=False),  # publish_artifacts
        ),
    )

    @pytest.mark.parametrize("cli_test_case", _BUILD_OPTIONS, ids=_cli_args_case_id)
    def test_mock_build_op_selection(
        self, mocked_runner: MockedRunner, cli_test_case: BuildFlagCase
    ) -> None:
        cli_flags, expected_select_args, expected_create_args, expected_publish_args = (
            cli_test_case
        )
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["build", *cli_flags, spec_path_to_mock])
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        expected_output_dir = "_artifacts"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        mocked_runner.assert_build_config(expected_build_dir, PackageIndexConfig())
        # Defaults to selecting operations rather than stacks
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_called_once_with(
            **expected_select_args
        )
        mocked_build_env.select_layers.assert_not_called()
        # Always creates the environments to perform the requested operations
        mocked_build_env.create_environments.assert_called_once_with(
            **expected_create_args
        )
        # "Disabling" artifact publication triggers a dry run rather than skipping it completely
        mocked_build_env.publish_artifacts.assert_called_once_with(
            expected_output_dir, **expected_publish_args
        )
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    # Specific CLI option handling test cases for the "lock" subcommand
    LockFlagCase = tuple[tuple[str, ...], dict[str, bool], dict[str, bool]]
    _LOCK_OPTIONS = (
        # Define how the various flags in the CLI lock subcommand map to API method options
        # `--include` and its related options are tested separately
        (
            (),
            dict(lock=True, build=False, publish=False),  # select_operations
            dict(clean=False),  # lock_environments
        ),
        (
            ("--clean",),
            dict(lock=True, build=False, publish=False),  # select_operations
            dict(clean=True),  # lock_environments
        ),
        (
            ("--no-clean",),
            dict(lock=True, build=False, publish=False),  # select_operations
            dict(clean=False),  # lock_environments
        ),
    )

    @pytest.mark.parametrize("cli_test_case", _LOCK_OPTIONS, ids=_cli_args_case_id)
    def test_mock_lock_op_selection(
        self, mocked_runner: MockedRunner, cli_test_case: LockFlagCase
    ) -> None:
        cli_flags, expected_select_args, expected_lock_args = cli_test_case
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["lock", *cli_flags, spec_path_to_mock])
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        mocked_runner.assert_build_config(expected_build_dir, PackageIndexConfig())
        # Defaults to selecting operations rather than stacks
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_called_once_with(
            **expected_select_args
        )
        mocked_build_env.select_layers.assert_not_called()
        # Only locks the environments without fully building them
        mocked_build_env.lock_environments.assert_called_once_with(**expected_lock_args)
        mocked_build_env.create_environments.assert_not_called()
        # The lock subcommand doesn't even attempt the publication step
        mocked_build_env.publish_artifacts.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @staticmethod
    def get_invalid_lock_options() -> Generator[str, None, None]:
        # List of flags to check was defined when 'lock' was extracted from 'build'
        invalid_flag_names = ("lock", "build", "publish", "tag-outputs")
        for name in invalid_flag_names:
            yield f"--{name}"
            yield f"--no-{name}"

    @pytest.mark.parametrize("invalid_flag", list(get_invalid_lock_options()))
    def test_mock_lock_usage_error(
        self, mocked_runner: MockedRunner, invalid_flag: str
    ) -> None:
        mocked_spec_path = "/no/such/path/spec"
        result = mocked_runner.invoke(["lock", invalid_flag, mocked_spec_path])
        # Should complain about the invalid flag
        assert result.stdout[: len(EXPECTED_USAGE_PREFIX)] == EXPECTED_USAGE_PREFIX
        assert "Try 'python -m venvstacks lock --help' for help." in result.stdout
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        mocked_stack_spec.define_build_environment.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert isinstance(result.exception, SystemExit), report_traceback(
            result.exception
        )
        assert result.exit_code == 2

    # Specific CLI option handling test cases for the "publish" subcommand
    PublishFlagCase = tuple[tuple[str, ...], dict[str, bool], dict[str, bool]]
    _PUBLISH_OPTIONS = (
        # Define how the various flags in the CLI lock subcommand map to API method options
        # `--include` and its related options are tested separately
        (
            (),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=False, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--force",),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--dry-run",),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(dry_run=True, tag_outputs=False),  # publish_artifacts
        ),
        (
            ("--tag-outputs",),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=False, tag_outputs=True),  # publish_artifacts
        ),
        (
            (
                "--no-force",
                "--no-dry-run",
                "--no-tag-outputs",
            ),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=False, tag_outputs=False),  # publish_artifacts
        ),
    )

    @pytest.mark.parametrize("cli_test_case", _PUBLISH_OPTIONS, ids=_cli_args_case_id)
    def test_mock_publish_op_selection(
        self, mocked_runner: MockedRunner, cli_test_case: PublishFlagCase
    ) -> None:
        cli_flags, expected_select_args, expected_publish_args = cli_test_case
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["publish", *cli_flags, spec_path_to_mock])
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        expected_output_dir = "_artifacts"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        mocked_runner.assert_build_config(
            expected_build_dir, PackageIndexConfig.disabled()
        )
        # Defaults to selecting operations rather than stacks
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_called_once_with(
            **expected_select_args
        )
        mocked_build_env.select_layers.assert_not_called()
        # The publish subcommand assumes the environments are already created
        mocked_build_env.create_environments.assert_not_called()
        # The publish subcommand always attempts to publish the artifacts
        mocked_build_env.publish_artifacts.assert_called_once_with(
            expected_output_dir, **expected_publish_args
        )
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @staticmethod
    def get_invalid_publish_options() -> Generator[str, None, None]:
        # List of flags to check was defined when 'publish' was extracted from 'build'
        invalid_flag_names = ("lock", "build", "publish", "clean")
        for name in invalid_flag_names:
            yield f"--{name}"
            yield f"--no-{name}"

    @pytest.mark.parametrize("invalid_flag", list(get_invalid_publish_options()))
    def test_mock_publish_usage_error(
        self, mocked_runner: MockedRunner, invalid_flag: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["publish", invalid_flag, spec_path_to_mock])
        # Should complain about the invalid flag
        assert result.stdout[: len(EXPECTED_USAGE_PREFIX)] == EXPECTED_USAGE_PREFIX
        assert "Try 'python -m venvstacks publish --help' for help." in result.stdout
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        mocked_stack_spec.define_build_environment.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert isinstance(result.exception, SystemExit), report_traceback(
            result.exception
        )
        assert result.exit_code == 2

    # Specific CLI option handling test cases for the "publish" subcommand
    ExportFlagCase = tuple[tuple[str, ...], dict[str, bool], dict[str, bool]]
    _EXPORT_OPTIONS = (
        # Define how the various flags in the CLI lock subcommand map to API method options
        # `--include` and its related options are tested separately
        (
            (),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=False),  # export_environments
        ),
        (
            ("--force",),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=True),  # export_environments
        ),
        (
            ("--dry-run",),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(dry_run=True),  # export_environments
        ),
        (
            (
                "--no-force",
                "--no-dry-run",
            ),
            dict(lock=False, build=False, publish=True),  # select_operations
            dict(force=False),  # export_environments
        ),
    )

    @pytest.mark.parametrize("cli_test_case", _EXPORT_OPTIONS, ids=_cli_args_case_id)
    def test_mock_export_op_selection(
        self, mocked_runner: MockedRunner, cli_test_case: ExportFlagCase
    ) -> None:
        cli_flags, expected_select_args, expected_publish_args = cli_test_case
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["local-export", *cli_flags, spec_path_to_mock])
        # Always loads the stack spec and creates the build environment
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        expected_build_dir = "_build"
        expected_output_dir = "_export"
        mocked_stack_spec.load.assert_called_once_with(spec_path_to_mock)
        mocked_runner.assert_build_config(
            expected_build_dir, PackageIndexConfig.disabled()
        )
        # Defaults to selecting operations rather than stacks
        mocked_build_env = mocked_runner.mocked_build_env
        mocked_build_env.select_operations.assert_called_once_with(
            **expected_select_args
        )
        mocked_build_env.select_layers.assert_not_called()
        # The publish subcommand assumes the environments are already created
        mocked_build_env.create_environments.assert_not_called()
        # The publish subcommand always attempts to publish the artifacts
        mocked_build_env.export_environments.assert_called_once_with(
            expected_output_dir, **expected_publish_args
        )
        # Check operation result last to ensure test results are as informative as possible
        assert result.exception is None, report_traceback(result.exception)
        assert result.exit_code == 0

    @staticmethod
    def get_invalid_export_options() -> Generator[str, None, None]:
        # List of flags to check was defined when 'local-export' was derived from 'publish'
        invalid_flag_names = ("lock", "build", "publish", "clean", "tag-outputs")
        for name in invalid_flag_names:
            yield f"--{name}"
            yield f"--no-{name}"

    @pytest.mark.parametrize("invalid_flag", list(get_invalid_export_options()))
    def test_mock_export_usage_error(
        self, mocked_runner: MockedRunner, invalid_flag: str
    ) -> None:
        spec_path_to_mock = "/no/such/path/spec"
        result = mocked_runner.invoke(["local-export", invalid_flag, spec_path_to_mock])
        # Should complain about the invalid flag
        assert result.stdout[: len(EXPECTED_USAGE_PREFIX)] == EXPECTED_USAGE_PREFIX
        assert (
            "Try 'python -m venvstacks local-export --help' for help." in result.stdout
        )
        # No stack spec should be created
        mocked_stack_spec = mocked_runner.mocked_stack_spec
        mocked_stack_spec.assert_not_called()
        mocked_stack_spec.load.assert_not_called()
        mocked_stack_spec.define_build_environment.assert_not_called()
        # Check operation result last to ensure test results are as informative as possible
        assert isinstance(result.exception, SystemExit), report_traceback(
            result.exception
        )
        assert result.exit_code == 2



"""Allow execution of the package as a script."""

from .cli import main

# Handle multiprocessing potentially re-running this module with a name other than `__main__`
if __name__ == "__main__":
    main()
